--Copyright 2023 Snap One, LLC. All rights reserved.

JSON = require ('drivers-common-public.module.json')

require ('drivers-common-public.global.lib')
require ('drivers-common-public.global.handlers')
require ('drivers-common-public.global.timer')

do	-- Globals
	ALL_LIGHTS_ON = ALL_LIGHTS_ON or '0'
	LIGHTS = LIGHTS or {}

	PRESET_SOURCE_COUNT = 5

	Timer = {}

	PRESET_SELECTED = 0

	IN_PRODUCTION = true

	Buttons = {
		[303] = {DO_PUSH = 'NUMBER_0', },
		[304] = {DO_PUSH = 'NUMBER_1', },
		[305] = {DO_PUSH = 'NUMBER_2', },
		[306] = {DO_PUSH = 'NUMBER_3', },
		[307] = {DO_PUSH = 'NUMBER_4', },
		[308] = {DO_PUSH = 'NUMBER_5', },
		[309] = {DO_PUSH = 'NUMBER_6', },
		[310] = {DO_PUSH = 'NUMBER_7', },
		[311] = {DO_PUSH = 'NUMBER_8', },
		[312] = {DO_PUSH = 'NUMBER_9', },
		[326] = {DO_PUSH = 'START_SCAN_FWD', DO_RELEASE = 'STOP_SCAN_FWD', DO_CLICK = 'STOP_SCAN_FWD', },
		[334] = {DO_PUSH = 'MUTE_ON', },
		[333] = {DO_PUSH = 'MUTE_OFF', },
		[335] = {DO_PUSH = 'MUTE_TOGGLE', },
		[344] = {DO_PUSH = 'PAUSE', },
		[346] = {DO_PUSH = 'PLAY', },
		[331] = {DO_PUSH = 'PLAYPAUSE', },
		[351] = {DO_PUSH = 'START_SCAN_REV', DO_RELEASE = 'STOP_SCAN_REV', DO_CLICK = 'STOP_SCAN_REV', },
		[352] = {DO_PUSH = 'ROOM_OFF', },
		[354] = {DO_PUSH = 'SKIP_REV', },
		[355] = {DO_PUSH = 'SKIP_FWD', },
		[356] = {DO_PUSH = 'STOP', },
		[358] = {DO_PUSH = 'START_VOL_DOWN', DO_RELEASE = 'STOP_VOL_DOWN', DO_CLICK = 'STOP_VOL_DOWN', },
		[359] = {DO_PUSH = 'START_VOL_UP', DO_RELEASE = 'STOP_VOL_UP', DO_CLICK = 'STOP_VOL_UP', },
		[317] = {DO_PUSH = 'START_CH_UP', DO_RELEASE = 'STOP_CH_UP', DO_CLICK = 'STOP_CH_UP', },
		[316] = {DO_PUSH = 'START_CH_DOWN', DO_RELEASE = 'STOP_CH_DOWN', DO_CLICK = 'STOP_CH_DOWN', },
		[342] = {DO_PUSH = 'START_PAGE_DOWN', DO_RELEASE = 'STOP_PAGE_DOWN', DO_CLICK = 'STOP_PAGE_DOWN', },
		[343] = {DO_PUSH = 'START_PAGE_UP', DO_RELEASE = 'STOP_PAGE_UP', DO_CLICK = 'STOP_PAGE_UP', },
		[380] = {DO_PUSH = 'AV AND LIGHTS OFF', }, -- needs catching as this isn't a valid room command
	}

	OtherButtons = {
		[300] = {DO_PUSH = 'DASH', },
		[301] = {DO_PUSH = 'POUND', },
		[302] = {DO_PUSH = 'STAR', },
		[313] = {DO_PUSH = 'BACK', },
		[314] = {DO_PUSH = 'PROGRAM_D', }, -- Blue
		[315] = {DO_PUSH = 'CANCEL', },
		[318] = {DO_PUSH = 'CLOSED_CAPTIONED', },
		[319] = {DO_PUSH = 'CONTROL4', },
		[320] = {DO_PUSH = 'CUSTOM_1', },
		[321] = {DO_PUSH = 'CUSTOM_2', },
		[322] = {DO_PUSH = 'CUSTOM_3', },
		[323] = {DO_PUSH = 'PVR', }, -- DVR
		[324] = {DO_PUSH = 'EJECT', },
		[325] = {DO_PUSH = 'EXIT', },
		[327] = {DO_PUSH = 'FUNCTION', },
		[328] = {DO_PUSH = 'PROGRAM_B', }, -- Green
		[329] = {DO_PUSH = 'GUIDE', },
		[330] = {DO_PUSH = 'INFO', },
		[332] = {DO_PUSH = 'MENU', },
		[336] = {DO_PUSH = 'DOWN', },
		[337] = {DO_PUSH = 'LEFT', },
		[338] = {DO_PUSH = 'RIGHT', },
		[339] = {DO_PUSH = 'UP', },
		[340] = {DO_PUSH = 'OPEN_CLOSE', },
		[341] = {DO_PUSH = 'ORDER', },
		[345] = {DO_PUSH = 'PIP', },
		[347] = {DO_PUSH = 'POWER', },
		[348] = {DO_PUSH = 'RECALL', },
		[349] = {DO_PUSH = 'RECORD', },
		[350] = {DO_PUSH = 'PROGRAM_A', }, -- RED
		[353] = {DO_PUSH = 'ENTER', }, -- SELECT
		[361] = {DO_PUSH = 'PROGRAM_C', }, --YELLOW
	}

	ROOM_VARS_TO_WATCH = {
		ROOM_VARS.CURRENT_SELECTED_DEVICE,
		ROOM_VARS.CURRENT_AUDIO_DEVICE,
		ROOM_VARS.CURRENT_VIDEO_DEVICE,
		ROOM_VARS.POWER_STATE,
		ROOM_VARS.LAST_DEVICE_GROUP,
		ROOM_VARS.PLAYING_AUDIO_DEVICE,
		ROOM_VARS.CURRENT_MEDIA,
	}

	COMMANDS = {
		['Preset Cycler'] = 501,
		['Preset Cycler (Includes Off)'] = 500,
		['Preset Cycler - Reverse'] = 503,
		['Preset Cycler - Reverse (Includes Off)'] = 502,
		['Source Preset 1'] = 511,
		['Source Preset 2'] = 512,
		['Source Preset 3'] = 513,
		['Source Preset 4'] = 514,
		['Source Preset 5'] = 515,
		['Lights On'] = 400,
		['Lights Off'] = 401,
		['Lights Toggle'] = 402,
		['Room Off'] = 352,
		['AV and Lights Off'] = 380,
	}

	for command, idBinding in pairs (COMMANDS) do
		command = string.gsub (command, '%s+', '_')
		EC [command] = function (tParams)
			ReceivedFromProxy (idBinding, 'DO_PUSH', {})
		end
	end

	PRESET_REMOVED_TEXT = 'Preset No Longer Available'
end

function EC.ADDNUMBERS (tParams)
	for i = 0, 9 do
		C4:AddDynamicBinding (303 + i, "CONTROL", true, 'Number ' .. i, "BUTTON_LINK", false, false)
		PersistData.bindings [303 + i] = 'Number ' .. i
	end
end

function EC.ADDOTHERS (tParams)
	for k, v in pairs (OtherButtons) do
		C4:AddDynamicBinding (k, "CONTROL", true, v.DO_PUSH, "BUTTON_LINK", false, false)
		PersistData.bindings [k] = v.DO_PUSH
	end
end

function EC.Sync_Source_Between_Rooms (tParams)
	local srcRoom = tonumber (tParams ['Source Room'])
	local includeOff = (tParams ['Allow Power Off'] == 'On')

	for destRoom in string.gmatch (tParams ['Destination Room(s)'], '(%d+)') do
		destRoom = tonumber (destRoom)
		SyncRoomToRoom (srcRoom, destRoom, includeOff)
	end
end

function EC.Control_Button_Link (tParams)
	local id = tonumber (tParams.Command)
	local action = tParams.Type

	if (id and (Buttons [id] or OtherButtons [id])) then
		local strCommand = 'DO_PUSH'
		if (action == 'Release') then
			strCommand = 'DO_RELEASE'
		end
		ReceivedFromProxy (id, strCommand, {})
		if (action == 'Tap') then
			ReceivedFromProxy (id, 'DO_RELEASE', {})
		end
	end
end

function EC.SET_USE_DEFAULT_VOLUMES (tParams)
	if (tParams.USE_DEFAULTS == 1) then
		SUPPRESS_VOLUME_DEFAULTS = CancelTimer ('SUPPRESS_VOLUME_DEFAULTS')
	elseif (tParams.USE_DEFAULTS == 0) then
		local _timer = function (timer)
			SUPPRESS_VOLUME_DEFAULTS = CancelTimer ('SUPPRESS_VOLUME_DEFAULTS')
		end
		SUPPRESS_VOLUME_DEFAULTS = SetTimer ('SUPPRESS_VOLUME_DEFAULTS', 5 * ONE_MINUTE, _timer)
	end
end

function OnDriverDestroyed ()
	KillAllTimers ()
end

function OnDriverInit ()
	PersistData = PersistData or {}

	PersistData.bindings = PersistData.bindings or {}

	C4_DIGITAL_AUDIO = next (C4:GetDevicesByC4iName ('control4_digitalaudio.c4i'))

	for k, v in pairs(PersistData.bindings) do
		C4:AddDynamicBinding (k, "CONTROL", true, v, "BUTTON_LINK", false, false)
	end
end

function OnDriverLateInit (reason)
	LAST_LATE_INIT_REASON = reason
	C4:RegisterSystemEvent (C4SystemEvents.OnPIP, 0)

	C4:RegisterSystemEvent (C4SystemEvents.OnMediaInfoAdded, 0)
	C4:RegisterSystemEvent (C4SystemEvents.OnMediaInfoModified, 0)
	C4:RegisterSystemEvent (C4SystemEvents.OnMediaRemovedFromDevice, 0)
	C4:RegisterSystemEvent (C4SystemEvents.OnMediaDataRemoved, 0)

	AddVariable ('Watch Device Selected', '0', 'BOOL', true, false)
	AddVariable ('Listen Device Selected', '0', 'BOOL', true, false)

	AddVariable ('Last Device Selected Was Listen', '0', 'BOOL', true, false)
	AddVariable ('Last Device Selected Was Watch', '0', 'BOOL', true, false)

	AddVariable ('Current Device Selected', '0', 'NUMBER', true, false)
	AddVariable ('Last Device Selected', '0', 'NUMBER', true, false)
	AddVariable ('Room Is Off', '1', 'BOOL', true, false)

	AddVariable ('Current Media Selected', '0', 'NUMBER', true, false)

	AddVariable ('Control Rooms', '', 'STRING', true, false)

	AddVariable ('Watch Default Volume', '0', 'NUMBER', false, false)
	AddVariable ('Listen Default Volume', '0', 'NUMBER', false, false)

	AddVariable ('Any Lights On', '0', 'BOOL', true, false)
	AddVariable ('All Lights On', '0', 'BOOL', true, false)

	RegisterVariableListener (C4_DIGITAL_AUDIO, DIGITAL_AUDIO_VARS.ROOM_MAP_INFO, OWVC.ParseRoomMapInfo)

	PersistData = PersistData or {}
	PersistData.PRESET_SOURCES = PersistData.PRESET_SOURCES or {}
	PersistData.PRESET_SOURCE_MEDIA = PersistData.PRESET_SOURCE_MEDIA or {}

	if (Properties ['Selected Control Room'] == '') then
		local roomId = C4:RoomGetId ()
		local roomName = C4:ListGetDeviceName (roomId)
		C4:RenameDevice (C4:GetDeviceID (), roomName.. ' - Control')
	end

	PRESET_SOURCES = PersistData.PRESET_SOURCES
	PRESET_SOURCE_MEDIA = PersistData.PRESET_SOURCE_MEDIA

	for k, _ in pairs (Properties) do
		OnPropertyChanged (k)
	end
end

function OPC.Debug_Mode (value)
	CancelTimer ('DEBUGPRINT')
	DEBUGPRINT = (value == 'On')

	if (DEBUGPRINT) then
		local _timer = function (timer)
			C4:UpdateProperty ('Debug Mode', 'Off')
			OnPropertyChanged ('Debug Mode')
		end
		SetTimer ('DEBUGPRINT', 36000000, _timer)
	end
end

function OPC.Driver_Version (value)
	C4:UpdateProperty ('Driver Version', C4:GetDriverConfigInfo ('version'))
end

function OPC.Select_Room (value)
	if (value == '') then
		local thisRoom = C4:RoomGetId ()
		C4:UpdateProperty ('Select Room', thisRoom)
		OnPropertyChanged ('Select Room')
		return
	end

	for _, roomId in ipairs (CONTROL_ROOMS or {}) do
		for _, var in ipairs (ROOM_VARS_TO_WATCH) do
			UnregisterVariableListener (roomId, var)
		end
	end

	CONTROL_ROOMS = {}
	for roomId in string.gmatch (value, '(%d+)') do
		roomId = tonumber (roomId)
		table.insert (CONTROL_ROOMS, roomId)
	end

	SetVariable ('Control Rooms', table.concat (CONTROL_ROOMS, ','))

	C4:SetPropertyAttribs ('Watch Default Volume', ((#CONTROL_ROOMS > 1) and 1) or 0)
	C4:SetPropertyAttribs ('Listen Default Volume', ((#CONTROL_ROOMS > 1) and 1) or 0)
	C4:SetPropertyAttribs ('Reset Cycler Position', ((#CONTROL_ROOMS > 1) and 1) or 0)

	local name = {}
	for _, roomId in ipairs (CONTROL_ROOMS) do
		table.insert (name, C4:ListGetDeviceName (roomId))
	end
	name = table.concat (name, ', ')
	C4:UpdateProperty ('Selected Control Room', name)

	local pauseChecks = false
	if (LAST_LATE_INIT_REASON == nil or LAST_LATE_INIT_REASON == 'DIT_UPDATING') then
		pauseChecks = true
	end
	RegisterListeners (pauseChecks)

	if (#CONTROL_ROOMS == 1) then
		SetTimer ('CheckRoom', 250)
	else
		SetTimer ('CheckAllRooms', 250)
	end
end

function OPC.Watch_Default_Volume (value)
	WATCH_DEFAULT_VOLUME = tonumber (value)
	SetVariable ('Watch Default Volume', WATCH_DEFAULT_VOLUME)
end

function OPC.Listen_Default_Volume (value)
	LISTEN_DEFAULT_VOLUME = tonumber (value)
	SetVariable ('Listen Default Volume', LISTEN_DEFAULT_VOLUME)
end

function OPC.Reset_Cycler_Position (value)
	if (value == 'Never') then
		RESET_CYCLER = nil
	elseif (value == 'On Source Change') then
		RESET_CYCLER = 'SOURCE'
	elseif (value == 'On Room Off') then
		RESET_CYCLER = 'ROOM_OFF'
	elseif (value == 'On Source Change Or Room Off') then
		RESET_CYCLER = 'SOURCE_AND_ROOM_OFF'
	else
		RESET_CYCLER = nil
	end
end

function OPC.Ignore_OSD_Selection (value)
	if (value == 'Never') then
		IGNORE_OSD = nil
	elseif (value == 'Always') then
		IGNORE_OSD = 'ALWAYS'
	elseif (value == 'For Default Volume') then
		IGNORE_OSD = 'VOLUME'
	else
		IGNORE_OSD = nil
	end
end

function OPC.Preset_Source (value, preset)
	value = tonumber (value) or 0

	PRESET_SOURCES [preset] = value

	C4:MediaSetDeviceContext (value)
	local stations = C4:MediaGetAllBroadcastAudio () or {}

	if (next (stations)) then
		C4:SetPropertyAttribs ('Preset Source ' .. preset .. ' Media', 0)

		local currentMediaId = PRESET_SOURCE_MEDIA [preset]
		local currentLine

		local list = {}

		for mediaId, stationId in pairs (stations or {}) do
			mediaId = tonumber (mediaId)
			local info = C4:MediaGetBroadcastAudioInfo (mediaId)
			local line = info.name .. ' [' .. mediaId .. ']'
			line = string.gsub (line, ',', ' ')

			table.insert (list, line)

			if (mediaId == currentMediaId) then
				currentLine = line
			end
		end
		if (currentMediaId and not currentLine) then
			table.insert (list, PRESET_REMOVED_TEXT)
			currentLine = PRESET_REMOVED_TEXT
			PRESET_SOURCE_MEDIA [preset] = -1
		end

		table.sort (list)
		list = table.concat (list, ',')

		C4:UpdatePropertyList ('Preset Source ' .. preset .. ' Media', list)

		if (currentLine) then
			--this is to force the redisplay of the already selected property as the list has refreshed
			C4:UpdateProperty ('Preset Source ' .. preset .. ' Media', '')
			C4:UpdateProperty ('Preset Source ' .. preset .. ' Media', currentLine)
		end

	else
		C4:SetPropertyAttribs ('Preset Source ' .. preset .. ' Media', 1)

		C4:UpdatePropertyList ('Preset Source ' .. preset .. ' Media', '')

		PRESET_SOURCE_MEDIA [preset] = nil
	end
end

function OPC.Preset_Source_Media (value, preset)
	local mediaId = string.match (value, '.*%[(%d+)%]$')
	mediaId = tonumber (mediaId)
	if (mediaId) then
		local oldMediaId = PRESET_SOURCE_MEDIA [preset]
		PRESET_SOURCE_MEDIA [preset] = mediaId

		if (oldMediaId == -1) then
			OPC.Preset_Source (PRESET_SOURCES [preset], preset)
		end
	end
end

function OVC.Watch_Default_Volume (value)
	WATCH_DEFAULT_VOLUME = tonumber (value)
	UpdateProperty ('Watch Default Volume', WATCH_DEFAULT_VOLUME)
end

function OVC.Listen_Default_Volume (value)
	LISTEN_DEFAULT_VOLUME = tonumber (value)
	UpdateProperty ('Listen Default Volume', LISTEN_DEFAULT_VOLUME)
end

function OWVC.ParseRoomMapInfo (idDevice, idVariable, strValue)
	local info = strValue or ''
	QueueMap = {}
	RoomQIDMap = {}

	for audioQueueInfo in string.gmatch (info, '<audioQueueInfo>(.-)</audioQueueInfo>') do
		local queue = string.match (audioQueueInfo, '<queue>(.-)</queue>')

		local qId = tonumber (string.match (queue, '<id>(.-)</id>'))

		local source = tonumber (string.match (queue, '<device_id>(.-)</device_id>'))
		local state = string.match (queue, '<state>(.-)</state>')
		local ownerId = tonumber (string.match (queue, '<owner>(.-)</owner>'))

		QueueMap [qId] = {
			source = source,
			state = state,
			ownerId = ownerId,
			qId = qId,
		}

		table.insert (QueueMap [qId], ownerId)

		local rooms = string.match (queue, '<rooms>(.-)</rooms>')

		for roomId in string.gmatch (rooms, '<id>(.-)</id>') do
			roomId = tonumber (roomId)
			if (roomId ~= ownerId) then
				table.insert (QueueMap [qId], roomId)
			end
			RoomQIDMap [roomId] = qId
		end
	end
end

function OWVC.LightsChanged (idDevice, idVariable, strValue)
	if (idVariable == 1000) then

		LIGHTS [idDevice] = strValue
		dbg ('LIGHTS ' .. idDevice .. ' is now ' .. strValue)

		SetTimer ('CheckStatus', 250)
	end
end

function OWVC.RoomVarChanged (idDevice, idVariable, strValue)
	if (PAUSE_CHECKS) then
		print ('Not running room AV checks')
		return
	end
	local isControlRoom = false
	for _, roomId in ipairs (CONTROL_ROOMS or {}) do
		if (roomId == idDevice) then
			isControlRoom = true
		end
	end

	if (isControlRoom) then
		if (#CONTROL_ROOMS == 1) then
			SetTimer ('CheckRoom', 100)
		else
			SetTimer ('CheckAllRooms', 100)
		end
	end
end

do -- setup individual handlers for each count
	for preset = 1, PRESET_SOURCE_COUNT do
		OPC ['Preset_Source_' .. tostring (preset)] = function (value)
			OPC.Preset_Source (value, preset)
		end
		OPC ['Preset_Source_' .. tostring (preset) .. '_Media'] = function (value)
			OPC.Preset_Source_Media (value, preset)
		end
	end
end

function OSE.OnPIP (event)
	local pauseChecks = true
	RegisterListeners (pauseChecks)
end

function CheckPresetMediaChanged (idDevice)
	for preset = 1, PRESET_SOURCE_COUNT do
		if (PRESET_SOURCES [preset] == idDevice) then
			OPC.Preset_Source (idDevice, preset)
		end
	end
end

function OSE.OnMediaInfoAdded (event)
	local idDevice = string.match (event, '<param name="iddevice" type="ulong">(%d+)</param>')
	idDevice = tonumber (idDevice)
	CheckPresetMediaChanged (idDevice)
end

function OSE.OnMediaInfoModified (event)
	local idDevice = string.match (event, '<param name="iddevice" type="ulong">(%d+)</param>')
	idDevice = tonumber (idDevice)
	CheckPresetMediaChanged (idDevice)
end

function OSE.OnMediaRemovedFromDevice (event)
	local idDevice = string.match (event, '<param name="iddevice" type="ulong">(%d+)</param>')
	idDevice = tonumber (idDevice)
	CheckPresetMediaChanged (idDevice)
end

function OSE.OnMediaDataRemoved (event)
	local idDevice = string.match (event, '<param name="iddevice" type="ulong">(%d+)</param>')
	idDevice = tonumber (idDevice)
	CheckPresetMediaChanged (idDevice)
end

function ReceivedFromProxy (idBinding, strCommand, tParams)
	dbg (idBinding .. ' command: ' .. strCommand)
	if (DEBUGPRINT) then
		Print (tParams)
	end

	-- Lights On
	if (idBinding == 400) then
		if (strCommand == 'DO_PUSH') then
			LightsOn ()
		end

	-- Lights Off
	elseif (idBinding == 401) then
		if (strCommand == 'DO_PUSH') then
			LightsOff ()
		end

	-- Lights Toggle
	elseif (idBinding == 402) then
		if (strCommand == 'DO_PUSH') then
			LightsToggle ()
		end

	--PLAYPAUSE
	elseif (idBinding == 331) then
		if (strCommand == 'DO_PUSH') then
			local roomsToPlayPause = {}
			local devicesToPlayPause = {}
			for _, roomId in ipairs (CONTROL_ROOMS) do
				local watch, listen, currentDevice, isDigitalMedia = GetWatchOrListen (roomId)

				local spotifyConnectDeviceIds = C4:GetDevicesByC4iName ('SpotifyConnect.c4z')

				for deviceId, _ in pairs (spotifyConnectDeviceIds) do
					local spotifyConnectProxyId = C4:GetProxyDevicesById (deviceId)
					if (currentDevice == spotifyConnectProxyId) then
						isDigitalMedia = false
					end
				end

				print (watch, listen, currentDevice, isDigitalMedia)

				if (isDigitalMedia) then
					local qId, q = GetQueueIDByRoomID (roomId)
					if (q.state == 'Play') then
						roomsToPlayPause [q.ownerId] = 'PAUSE'
					else
						roomsToPlayPause [q.ownerId] = 'PLAY'
					end

				elseif (currentDevice ~= 0) then
					devicesToPlayPause [currentDevice] = true
				end
			end

			for ownerId, cmd in pairs (roomsToPlayPause) do
				C4:SendToDevice (ownerId, cmd, {})
			end

			for deviceId, _ in pairs (devicesToPlayPause) do
				C4:SendToDevice (deviceId, 'PLAYPAUSE', {})
			end
		end

	--AV AND LIGHTS OFF
	elseif (idBinding == 380) then
		if (strCommand == 'DO_PUSH') then
			LightsOff ()
			TurnOffRooms ()
		end

	elseif (idBinding == 500 or idBinding == 501 or idBinding == 502 or idBinding == 503) then

		if (strCommand == 'DO_PUSH') then

			local highestPreset = 0
			for preset = 1, PRESET_SOURCE_COUNT do
				if (PRESET_SOURCES [preset] and PRESET_SOURCES [preset] ~= 0) then
					highestPreset = preset
				end
			end

			if (highestPreset == 0) then
				print ('nothing to cycle')
				return
			end

			local includeOff = idBinding == 500 or idBinding == 502
			local direction = ((idBinding == 500 or idBinding == 501) and 1) or ((idBinding == 502 or idBinding == 503) and -1)

			local index = PRESET_SELECTED

			if (index == 0 and direction == -1) then
				index = highestPreset + 1
			end

			repeat
				index = index + direction
				if (index > highestPreset) then
					index = 0
					if (includeOff) then
						TurnOffRooms ()
						PRESET_SELECTED = index
						break
					end

				elseif (index < 1) then
					index = 0
					if (includeOff) then
						TurnOffRooms ()
						PRESET_SELECTED = index
						break
					else
						index = highestPreset + 1
					end

				else
					if (PRESET_SOURCES [index] and PRESET_SOURCES [index] ~= 0) then
						SelectPreset (index)
						break
					end
				end

			until index == PRESET_SELECTED
		end

	elseif (idBinding >= 511 and idBinding <= 515) then
		if (strCommand == 'DO_PUSH') then
			local preset = idBinding - 510
			SelectPreset (preset)
		end

	else
		local cmd
		if (Buttons [idBinding]) then
			if (Buttons [idBinding] [strCommand]) then
				cmd = Buttons [idBinding] [strCommand]
			end
		elseif (OtherButtons [idBinding]) then
			if (OtherButtons [idBinding] [strCommand]) then
				cmd = OtherButtons [idBinding] [strCommand]
			end
		end
		if (cmd) then
			for _, roomId in ipairs (CONTROL_ROOMS) do
				local send = true
				if (cmd == 'ROOM_OFF') then
					send = (C4:GetDeviceVariable (roomId, ROOM_VARS.POWER_STATE) == '1')
				end
				if (send) then
					C4:SendToDevice (roomId, cmd, {})
				end
			end
		end
	end
end

function RegisterListeners (pauseChecks)
	PAUSE_CHECKS = pauseChecks
	for deviceId, _ in pairs (LIGHTS or {}) do
		UnregisterVariableListener (deviceId, 1000)
	end
	LIGHTS = {}
	for _, roomId in ipairs (CONTROL_ROOMS) do
		local lights = C4:SendToDevice (roomId , "GET_LIGHT_DEVICES", {}) or ''
		for source in string.gmatch (lights, "<source>(.-)</source>") do
			local id = tonumber (string.match (source, "<id>(.-)</id>")) or 0
			LIGHTS [id] = C4:GetDeviceVariable (id, 1000)
			RegisterVariableListener (id, 1000, OWVC.LightsChanged)
		end

		for _, var in ipairs (ROOM_VARS_TO_WATCH) do
			RegisterVariableListener (roomId, var, OWVC.RoomVarChanged)
		end
	end
	PAUSE_CHECKS = nil
end

function LightsOn ()
	ALL_LIGHTS_ON = '1'
	for light, state in pairs (LIGHTS) do
		C4:SendToDevice (light, "ON", {})
	end
	C4:SendToProxy (400, 'MATCH_LED_STATE', {STATE = '1'})
	C4:SendToProxy (401, 'MATCH_LED_STATE', {STATE = '0'})
	C4:SendToProxy (402, 'MATCH_LED_STATE', {STATE = '1'})
end

function LightsOff ()
	ALL_LIGHTS_ON = '0'
	for light, state in pairs (LIGHTS) do
		C4:SendToDevice (light, "OFF", {})
	end
	C4:SendToProxy (400, 'MATCH_LED_STATE', {STATE = '0'})
	C4:SendToProxy (401, 'MATCH_LED_STATE', {STATE = '1'})
	C4:SendToProxy (402, 'MATCH_LED_STATE', {STATE = '0'})
end

function LightsToggle ()
	if (ALL_LIGHTS_ON == '1') then
		LightsOff ()
	else
		LightsOn ()
	end
end

function SelectPreset (index)
	local source = PRESET_SOURCES [index]
	if (source) then
		if (index == PRESET_SELECTED) then
			local _, _, currentDevice = GetWatchOrListen (CONTROL_ROOMS [1])
			if (currentDevice == PRESET_SOURCES [index]) then
				--[[
					assume that if:
						- we're already on the right source for the room
						- we're reselecting the same index
					then it shouldn't fire
				]]
				return
			end
		end

		SetCurrentPreset (index)

		if (PRESET_SOURCE_MEDIA [index]) then
			local roomId = CONTROL_ROOMS [1]
			local mediaId = PRESET_SOURCE_MEDIA [index]

			local addRooms = {}
			for i = 2, #CONTROL_ROOMS do
				table.insert (addRooms, CONTROL_ROOMS [i])
			end
			if (#addRooms > 0) then
				addRooms = table.concat (addRooms, ',')
			else
				addRooms = nil
			end

			local args = {
				deselect = '0',
				type = 'BROADCAST_AUDIO',
				mediaid = mediaId,
			}
			C4:SendToDevice (roomId, 'SELECT_AUDIO_MEDIA', args)

			if (addRooms) then
				local args = {
					ROOM_ID = roomId,
					ROOM_ID_LIST = addRooms,
				}
				C4:SendToDevice (C4_DIGITAL_AUDIO, 'ADD_ROOMS_TO_SESSION', args)
			end

		else
			for _, roomId in ipairs (CONTROL_ROOMS or {}) do
				local deviceIds = {}

				local watchDevices, watchDevicesIds = SourceIdParser (C4:SendToDevice (roomId, 'GET_WATCH_DEVICES', {}))
				local watchDevicesHidden, watchDevicesHiddenIds = SourceIdParser (C4:SendToDevice (roomId, 'GET_WATCH_DEVICES', {hidden = 1}))

				local listenDevices, listenDevicesIds = SourceIdParser (C4:SendToDevice (roomId, 'GET_LISTEN_DEVICES', {}))
				local listenDevicesHidden, listenDevicesHiddenIds = SourceIdParser (C4:SendToDevice (roomId, 'GET_LISTEN_DEVICES', {hidden = 1}))

				table.insert (deviceIds, watchDevicesIds)
				table.insert (deviceIds, watchDevicesHiddenIds)
				table.insert (deviceIds, listenDevicesIds)
				table.insert (deviceIds, listenDevicesHiddenIds)

				for tab, devices in ipairs (deviceIds) do
					if (devices [source]) then
						local command = ((tab == 1 or tab == 2) and 'SELECT_VIDEO_DEVICE') or ((tab == 3 or tab == 4) and 'SELECT_AUDIO_DEVICE') or nil
						local is_digital_audio = (string.lower (C4:GetDeviceData (source, 'digital_audio_support') or '') == 'true')
						if (command) then
							if (command == 'SELECT_AUDIO_DEVICE' and is_digital_audio) then
								C4:SendToDevice (source, 'DEVICE_SELECTED', {idRoom = roomId})
							else
								C4:SendToDevice (roomId, command, {deviceid = source})
							end
						end
						break
					end
				end
			end
		end
	end
end

function SetCurrentPreset (index)
	if (PRESET_SELECTED ~= index) then
		PRESET_SELECTED = index
		C4:FireEvent ('Preset ' .. index .. ' Selected')
	end
end

function TurnOffRooms ()
	for _, roomId in ipairs (CONTROL_ROOMS or {}) do
		local send = (C4:GetDeviceVariable (roomId, ROOM_VARS.POWER_STATE) == '1')
		if (send) then
			C4:SendToDevice (roomId, 'ROOM_OFF', {})
		end
	end
end

function CheckStatus ()
	dbg ('Checking Status')
	local on = '1'
	local off = '1'

	local anyOn = '0'
	local allOn = '1'

	for id, status in pairs (LIGHTS) do
		dbg (id .. ' - ' .. status)
		if (status == '0') then
			on = '0'
			allOn = '0'
		else
			off = '0'
			anyOn = '1'
		end
	end

	if (ALL_LIGHTS_ON ~= on) then
		ALL_LIGHTS_ON = on

		C4:SendToProxy (400, 'MATCH_LED_STATE', {STATE = on})
		C4:SendToProxy (401, 'MATCH_LED_STATE', {STATE = off})
		C4:SendToProxy (402, 'MATCH_LED_STATE', {STATE = on})
	end

	if (Variables ['Any Lights On'] ~= anyOn) then
		dbg ('Any Lights On' .. ' - ' .. anyOn)
		SetVariable ('Any Lights On', anyOn)
	end
	if (Variables ['All Lights On'] ~= allOn) then
		dbg ('All Lights On' .. ' - ' .. allOn)
		SetVariable ('All Lights On', allOn)
	end
end

function GetWatchOrListen (roomId)
	local currentDevice = tonumber (C4:GetDeviceVariable (roomId, ROOM_VARS.CURRENT_SELECTED_DEVICE)) or 0
	local currentAudioDevice = tonumber (C4:GetDeviceVariable (roomId, ROOM_VARS.CURRENT_AUDIO_DEVICE)) or 0
	local currentVideoDevice = tonumber (C4:GetDeviceVariable (roomId, ROOM_VARS.CURRENT_VIDEO_DEVICE)) or 0
	local powerState = (C4:GetDeviceVariable (roomId, ROOM_VARS.POWER_STATE) == '1')
	local deviceGroup = C4:GetDeviceVariable (roomId, ROOM_VARS.LAST_DEVICE_GROUP)
	local playingAudioDevice = tonumber (C4:GetDeviceVariable (roomId, ROOM_VARS.PLAYING_AUDIO_DEVICE)) or 0

	local isDigitalMedia = (currentDevice == C4_DIGITAL_AUDIO)

	-- catch Digital Media
	if (isDigitalMedia) then
		if (playingAudioDevice and playingAudioDevice ~= 0) then
			currentDevice = playingAudioDevice
			currentAudioDevice = playingAudioDevice
		else
			local qId, q = GetQueueIDByRoomID (roomId)

			if (q and q.source) then
				currentDevice = q.source
				currentAudioDevice = q.source
			else
				--  digital media, not in a queue, not a master /shrug
			end
		end
	end

	local listen, watch

	if (currentDevice == 0 or powerState == false) then
		listen = false
		watch = false

	elseif (currentAudioDevice ~= currentVideoDevice) then
		listen = (currentDevice == currentAudioDevice)
		watch = (currentDevice == currentVideoDevice)

	elseif (currentAudioDevice == currentVideoDevice) then
		listen = (deviceGroup == 'listen' or deviceGroup == 'audio')
		watch = (deviceGroup == 'watch' or deviceGroup == 'video')
	end

	local osdDevice = C4:GetBoundProviderDevice (roomId, 6)
	local inNavigation = (osdDevice ~= 0 and currentDevice == osdDevice)

	return watch, listen, currentDevice, isDigitalMedia, inNavigation
end

function CheckRoom ()
	local roomId = CONTROL_ROOMS [1]

	local watch, listen, currentDevice, isDigitalMedia, inNavigation = GetWatchOrListen (roomId)
	local osdDevice = C4:GetBoundProviderDevice (roomId, 6)

	if (inNavigation and IGNORE_OSD == 'ALWAYS') then
		return
	end

	local currentMediaId = C4:GetDeviceVariable (roomId, ROOM_VARS.CURRENT_MEDIA)
	currentMediaId = string.match (currentMediaId, '<mediaid>(%d+)</mediaid>')
	currentMediaId = tonumber (currentMediaId) or 0

	local oldMediaId = tonumber (Variables ['Current Media Selected'])

	local oldDevice = tonumber (Variables ['Current Device Selected'])

	local currentPreset

	if (currentDevice ~= oldDevice) then
		SetVariable ('Last Device Selected', oldDevice or '0')
		SetVariable ('Last Device Selected Was Listen', Variables ['Listen Device Selected'] or '0')
		SetVariable ('Last Device Selected Was Watch', Variables ['Watch Device Selected'] or '0')

		SetVariable ('Current Device Selected', currentDevice)

		if (oldDevice == 0) then
			C4:FireEvent ('Room Turned On')
		end

		if (RESET_CYCLER) then
			if (currentDevice == 0) then
				if (RESET_CYCLER == 'ROOM_OFF' or RESET_CYCLER == 'SOURCE_AND_ROOM_OFF') then
					currentPreset = 0

				end
			else
				if (RESET_CYCLER == 'SOURCE' or RESET_CYCLER == 'SOURCE_AND_ROOM_OFF') then
					currentPreset = 0
					for index, source in ipairs (PRESET_SOURCES) do
						if (source == currentDevice) then
							if (PRESET_SOURCE_MEDIA [index] == nil) then
								currentPreset = index
								break
							else
								currentPreset = nil
							end
						end
					end
				end
			end
		end
	end

	if (currentMediaId ~= oldMediaId) then
		SetVariable ('Current Media Selected', currentMediaId)
		if (currentDevice ~= 0 and currentMediaId ~= 0) then
			if (RESET_CYCLER == 'SOURCE' or RESET_CYCLER == 'SOURCE_AND_ROOM_OFF') then
				currentPreset = 0
				for index, mediaId in pairs (PRESET_SOURCE_MEDIA) do
					if (mediaId == currentMediaId) then
						currentPreset = index
						break
					end
				end
			end
		end

	elseif (currentMediaId ~= 0) then
		currentPreset = nil
	end

	if (currentPreset) then
		SetCurrentPreset (currentPreset)
	end

	if (watch) then
		C4:FireEvent ('Watch Device Selected')
		if (Variables ['Watch Device Selected'] == '0' or (oldDevice == osdDevice and (IGNORE_OSD == 'VOLUME' or IGNORE_OSD == 'ALWAYS'))) then
			if (Variables ['Watch Device Selected'] == '0') then
				C4:FireEvent ('Entered Watch Mode')
			end
			if (not (SUPPRESS_VOLUME_DEFAULTS) and WATCH_DEFAULT_VOLUME > 0 and C4:GetDeviceVariable (roomId, ROOM_VARS.HAS_DISCRETE_VOLUME)) then
				if (inNavigation and (IGNORE_OSD == 'VOLUME' or IGNORE_OSD == 'ALWAYS')) then
				else
					C4:SendToDevice (roomId, 'SET_VOLUME_LEVEL', {LEVEL = WATCH_DEFAULT_VOLUME})
				end
			end
		end

		if (Variables ['Listen Device Selected'] == '1') then
			C4:FireEvent ('Left Listen Mode')
		end

	elseif (listen) then
		C4:FireEvent ('Listen Device Selected')
		if (Variables ['Listen Device Selected'] == '0' or (oldDevice == osdDevice and (IGNORE_OSD == 'VOLUME' or IGNORE_OSD == 'ALWAYS'))) then
			if (Variables ['Listen Device Selected'] == '0') then
				C4:FireEvent ('Entered Listen Mode')
			end
			if (not (SUPPRESS_VOLUME_DEFAULTS) and LISTEN_DEFAULT_VOLUME > 0 and C4:GetDeviceVariable (roomId, ROOM_VARS.HAS_DISCRETE_VOLUME)) then
				if (inNavigation and (IGNORE_OSD == 'VOLUME' or IGNORE_OSD == 'ALWAYS')) then
				else
					C4:SendToDevice (roomId, 'SET_VOLUME_LEVEL', {LEVEL = LISTEN_DEFAULT_VOLUME})
				end
			end
		end

		if (Variables ['Watch Device Selected'] == '1') then
			C4:FireEvent ('Left Watch Mode')
		end

	else
		C4:FireEvent ('Room Turned Off')
		if (Variables ['Watch Device Selected'] == '1') then
			C4:FireEvent ('Left Watch Mode')
		end
		if (Variables ['Listen Device Selected'] == '1') then
			C4:FireEvent ('Left Listen Mode')
		end

	end

	SetVariable ('Room Is Off', ((watch or listen) and '0') or '1')
	SetVariable ('Watch Device Selected', (watch and '1') or '0')
	SetVariable ('Listen Device Selected', (listen and '1') or '0')
end

function CheckAllRooms ()
	local anyWatch, anyListen = false, false
	for _, roomId in ipairs (CONTROL_ROOMS or {}) do
		local watch, listen, currentDevice, isDigitalMedia, inNavigation = GetWatchOrListen (roomId)

		if (inNavigation and IGNORE_OSD == 'ALWAYS') then
		else
			anyWatch = watch or anyWatch
			anyListen = listen or anyListen
		end
	end

	SetVariable ('Room Is Off', ((anyWatch or anyListen) and '0') or '1')
	SetVariable ('Watch Device Selected', (anyWatch and '1') or '0')
	SetVariable ('Listen Device Selected', (anyListen and '1') or '0')
end

function SyncRoomToRoom (srcRoom, destRoom, includeOff)
	local watch, listen, currentDevice, isDigitalMedia, inNavigation = GetWatchOrListen (srcRoom)

	if (inNavigation) then
		local currentAudioDevice = tonumber (C4:GetDeviceVariable (srcRoom, ROOM_VARS.CURRENT_AUDIO_DEVICE)) or 0
		isDigitalMedia = (currentAudioDevice == C4_DIGITAL_AUDIO)
	end

	if (isDigitalMedia) then
		local qId = GetQueueIDByRoomID (srcRoom)
		JoinRoomToSession (destRoom, qId)

	else
		if (watch) then
			local watchDevices, watchDevicesIds = SourceIdParser (C4:SendToDevice (destRoom, 'GET_WATCH_DEVICES', {}))
			local watchDevicesHidden, watchDevicesHiddenIds = SourceIdParser (C4:SendToDevice (destRoom, 'GET_WATCH_DEVICES', {hidden = 1}))

			local listenDevices, listenDevicesIds = SourceIdParser (C4:SendToDevice (destRoom, 'GET_LISTEN_DEVICES', {}))
			local listenDevicesHidden, listenDevicesHiddenIds = SourceIdParser (C4:SendToDevice (destRoom, 'GET_LISTEN_DEVICES', {hidden = 1}))

			if ((watchDevicesIds and watchDevicesIds [currentDevice]) or (watchDevicesHiddenIds and watchDevicesHiddenIds [currentDevice])) then
				C4:SendToDevice (destRoom, 'SELECT_VIDEO_DEVICE', {deviceid = currentDevice})

			elseif ((listenDevicesIds and listenDevicesIds [currentDevice]) or (listenDevicesHiddenIds and listenDevicesHiddenIds [currentDevice])) then
				C4:SendToDevice (destRoom, 'SELECT_AUDIO_DEVICE', {deviceid = currentDevice})
			end

		elseif (listen) then
			C4:SendToDevice (destRoom, 'SELECT_AUDIO_DEVICE', {deviceid = currentDevice})

		elseif (includeOff) then
			C4:SendToDevice (destRoom, 'ROOM_OFF', {})
		end
	end
end

function GetQueueIDByRoomID (roomId)
	roomId = tonumber (roomId)
	local qId = RoomQIDMap [roomId] or 0
	local queue = QueueMap [qId] or {}
	return qId, queue
end

function GetRoomMapByQueueID (qId)
	qId = tonumber (qId)
	local queue = QueueMap [qId] or {}
	return (queue)
end

function JoinRoomToSession (roomId, qId)
	local sessionQueue = GetRoomMapByQueueID (qId)
	local _, roomQueue = GetQueueIDByRoomID (roomId)

	if (sessionQueue.ownerId and sessionQueue.ownerId ~= roomQueue.ownerId) then
		local args = {
			ROOM_ID = sessionQueue.ownerId,
			ROOM_ID_LIST = roomId,
		}
		C4:SendToDevice (C4_DIGITAL_AUDIO, 'ADD_ROOMS_TO_SESSION', args)
		return true
	end
end

function SourceIdParser (strXML)
	local sources = {}
	local ids = {}
	for source in string.gmatch (strXML, '<source>(.-)</source>') do
		local id = string.match (source, '<id>(.-)</id>')
		id = tonumber (id)
		table.insert (sources, id)
		ids [id] = #sources
	end
	return sources, ids
end

function GetCommandList (currentValue, callbackWhenDone, search, filter)
	local list = {}

	for id, tab in pairs (Buttons) do
		local item = {
			value = id,
			text = string.gsub (tab.DO_PUSH, 'START_', ''),
		}
		table.insert (list, item)
	end
	for id, tab in pairs (OtherButtons) do
		local item = {
			value = id,
			text = string.gsub (tab.DO_PUSH, 'START_', ''),
		}
		table.insert (list, item)
	end

	local _sort = function (a, b)
		return (a.text < b.text)
	end

	table.sort (list, _sort)

	return list
end
