# Room Control

## Overview

This driver provides a convenient, connection-oriented method to avoid common room/keypad programming tasks by connecting a keypad button directly to a Room driver command.  Some additional “macro” capability has been added for certain options.

Additionally, programming events and variables are exposed to make conditional programming much simpler.

- Requires C4 OS 2.9.1 or later
- Works with any device that has a BUTTON_LINK input (Keypad, Keypad Dimmer, etc)

Some behaviors of this driver only work when there is only one room selected for this driver to control.  Those behaviors are marked with {SINGLE} where they appear in this document.

## Known Issues

As of version 110, the driver doesn’t have any means to specify the colors used for the BUTTON_LINK connections, and only the ALL LIGHTS ON/OFF/TOGGLE connections support feedback.

## Quick Configuration

Add the driver to the project in the room that you wish to control (this will by default be set as the room to control).

If you want to control additional rooms, add additional copies of the driver.

Each driver can also be set to control multiple rooms, with some limitations (see later)

Make connections to the BUTTON_LINK outputs as appropriate.

## Driver Properties

### Driver Version – [READ ONLY]

Shows the version of the driver as reported by Director

### Debug Mode – [On | Off]

Controls whether debug output is printed on the Lua tab in Composer Pro.

### Selected Control Room – [READ ONLY]

Shows the name(s) of the room(s) selected from the Select Room property.

### Select Room – [Multiselect Device Selector]

Selects the room(s) that this driver should control.  See notes in Connections, Driver Variables and Driver Events sections for special conditions when more than one room is selected.

### Ignore OSD Selection – [Always | For Default Volume | *Never*]

If set to Always, then no events or effects are triggered by the selection of the OSD in any monitored room.  If set to "For Default Volume", then only the Watch and Listen Default Volume effects are ignored and all other effects trigger as usual.  If set to Never, the OSD is treated like any other source.

### Watch Default Volume – [0 – 100] {SINGLE}

Whenever the Entered Watch Mode event is fired, the room is set to the volume level defined in this linked property/variable.  0 means do not set a volume.

### Listen Default Volume – [0 – 100] {SINGLE}

Whenever the Entered Listen Mode event is fired, the room is set to the volume level defined in this linked property/variable.  0 means do not set a volume.

### Preset Source 1-5 – [Single Device Selector]

Chooses the device that will be selected for this preset index, either from the direct PRESET SELECT X BUTTON_LINK connections, or when cycling through the presets with the PRESET CYCLER BUTTON_LINK connections.

### Preset Source 1-5 Media – [Dynamic List]

This presents a dynamic list of all media found in the Broadcast Audio section of the Control4 Media Database for the associated source (not all drivers populate any data into Broadcast Audio).  If an item is selected from this list, that media item will be selected as the source for the room(s) instead of the device itself, as per the Preset Source 1-5 property.

If no media is found then this property will not be visible.  If the selected device doesn't support Broadcast Audio selection then the programming event “When Preset x is selected“ should be used to trigger the desired selection in Composer programming.

### Reset Cycler Position– [Never | On Source Change | On Room Off | On Source Change Or Room Off] {SINGLE}

- Never: the internal counter that keeps track of the last source selected through the cycler is not affected by any action other than the PRESET CYCLER and PRESET CYCLER (INCLUDES OFF) BUTTON_LINK connections.
- On Source Change: selecting a source not on the Preset Source list will reset the counter to the beginning, and selecting a source on the Preset Source list will jump the counter to that Preset Source position.
- On Room Off: turning off the room will reset the counter to the beginning.
- On Source Change Or Room Off: combines the behaviors of On Source Change and On Room Off

## Driver Actions

### Add Number Connections

Adds the NUMBER_0 through NUMBER_9 BUTTON_LINK connections.

### Add Other Connections

Adds BUTTON_LINK connections for less commonly used Room driver commands  (see Connections section)

### Control Button Link

Emulates a tap (sends DO_PUSH then DO_CLICK), press (sends DO_PUSH) or release (sends DO_RELEASE) of a specified button (select from the …) list.

## Connections

There are BUTTON_LINK connections for the most popular room control connections “out of the box”:

- CH_UP
- CH_DOWN
- MUTE_ON
- MUTE_OFF
- MUTE_TOGGLE
- PAGE_DOWN
- PAGE_UP
- PAUSE
- PLAY
- *ROOM_OFF*
- SCAN_FWD
- SCAN_REV
- SKIP_REV
- SKIP_FWD
- STOP
- VOL_DOWN
- VOL_UP

The VOL_UP/DOWN, CH_UP/DOWN, PAGE_UP/DOWN and SCAN_FWD/REV are all set up to allow press and hold to send the expected START and STOP commands to the room(s).  All other commands are just single press actions.

*NOTE*: ROOM_OFF will only send the ROOM_OFF command to the selected control room(s) if the room is reporting as turned on.  This is to prevent ROOM_OFF programming from continually cycling.

The Add Number Connections Composer Action will add BUTTON_LINK connections for NUMBER_0 through NUMBER_9

The Add Other Connections Composer action will add additional BUTTON_LINK connections:

- DASH
- POUND
- STAR
- BACK
- BLUE
- CANCEL
- CLOSED_CAPTIONED
- CONTROL4
- CUSTOM_1
- CUSTOM_2
- CUSTOM_3
- DVR
- EJECT
- EXIT
- FUNCTION
- GREEN
- GUIDE
- INFO
- MENU
- NAVIGATE_DOWN
- NAVIGATE_LEFT
- NAVIGATE_RIGHT
- NAVIGATE_UP
- OPEN_CLOSE
- ORDER
- PIP
- POWER
- RECALL
- RECORD
- RED
- SELECT

### Special Connections

#### ALL LIGHTS ON

Turns on all the lights set in Composer to appear under Light.  The feedback for this BUTTON_LINK will turn on if all lights are on and off if any lights are off.

#### ALL LIGHTS OFF

Turns off all the lights set in Composer to appear under Light.  The feedback for this button will turn on if all lights are off and off in any lights are on.

#### ALL LIGHTS TOGGLE

If any of the lights in the room is Off, will turn on all lights, otherwise will turn off all lights.  The feedback for this button will turn on if all lights are on and off if any lights are off.

#### PLAYPAUSE

Send the PLAYPAUSE command through to the room and on to the current source driver, which must have a PLAYPAUSE command defined (this is not always standard).  If the current source in the room is a native Control4 MSP music service driver, this driver determines whether to send a PLAY or PAUSE based on the actual current state of the service.  Even if multiple rooms on this driver are using the same source device, only one command will be sent to that source device.

#### AV AND LIGHTS OFF

This macros the ROOM_OFF and ALL_LIGHTS_OFF function under one BUTTON_LINK connection.

#### PRESET CYCLER

Cycles through the sources listed in properties Preset Source 1-5, wrapping around from 5 to 1.  If no source is defined for a property, that index is skipped.

#### PRESET CYCLER (INCLUDES OFF)

Cycles through the sources listed in properties Preset Source 1-5, wrapping around from 5 to turning off the room, then on to 1.  If no source is defined for a property, that index is skipped.

## Driver Variables

The following variables are available for use in programming:

### Watch Device Selected - [BOOLEAN]

The source currently selected in this room (or any of the rooms) was selected through the Watch menu.

### Listen Device Selected - [BOOLEAN]

The source currently selected in this room (or any of the rooms) was selected through the Listen menu.

### Last Device Selected Was Listen - [BOOLEAN] {SINGLE}

The source selected before the current source in this room was selected through the Listen menu.

### Last Device Selected Was Watch - [BOOLEAN] {SINGLE}

The source selected before the current source in this room was selected through the Watch menu.

### Current Device Selected - [NUMBER]

The device ID of the currently selected source [only available if only one room is selected].

### Last Device Selected – [NUMBER] {SINGLE}

The device ID of the source selected before the current source in this room.

### Room Is Off – [BOOLEAN]

True if the room (or all the rooms) are off.

### Current Media Selected - [NUMBER]

The media ID of the last selected media.

### Watch Default Volume – [0 – 100] {SINGLE}

Whenever the Entered Watch Mode event is fired, the room is set to the volume level defined in this linked property/variable.  0 means do not set a volume.

### Listen Default Volume – [0 – 100] {SINGLE}

Whenever the Entered Listen Mode event is fired, the room is set to the volume level defined in this linked property/variable.  0 means do not set a volume.

### Any Lights On - [BOOLEAN]

Set to true if any of the lights controllable by this driver are On

### All Lights On - [BOOLEAN]

Set to true if all of the lights controllable by this driver are On

## Driver Events

### Watch Device Selected {SINGLE}

A device was selected from the Watch menu.

### Listen Device Selected {SINGLE}

A device was selected from the Listen menu.

### Entered Watch Mode {SINGLE}

A device was selected from the Watch menu, and either the previous device in the room was selected from the Listen menu or the room was off.  Same trigger as Default Watch Volume.

### Entered Listen Mode {SINGLE}

A device was selected from the Listen menu, and either the previous device in the room was selected from the Watch menu or the room was off.  Same trigger as Default Listen Volume.

### Room Turned Off {SINGLE}

The room turned off.

### Room Turned On {SINGLE}

The room turned on.

### Left Watch Mode {SINGLE}

A device was selected from the Listen menu or the room turned off, and the previous device in the room was selected from the Watch menu.

### Left Listen Mode {SINGLE}

A device was selected from the Watch menu or the room turned off, and the previous device in the room was selected from the Listen menu.

### Preset X Selected

The named preset was selected.

## Driver Commands

### Preset Cycler

Trigger Preset Source Cycler

### Preset Cycler (Includes Off)

Trigger Preset Source Cycler (Includes Off)

### Preset Cycler - Reverse

Trigger Preset Source Cycler - Reverse

### Preset Cycler - Reverse (Includes Off

Trigger Preset Source Cycler - Reverse (Includes Off)

### Source Preset 1

Trigger Source Preset 1

### Source Preset 2

Trigger Source Preset 2

### Source Preset 3

Trigger Source Preset 3

### Source Preset 4

Trigger Source Preset 4

### Source Preset 5

Trigger Source Preset 5

### Lights On

Trigger Lights On

### Lights Off

Trigger Lights Off

### Lights Toggle

Trigger Lights Toggle

### Room Off

Trigger Room Off

### AV and Lights Off

Trigger AV and Lights Off

### Sync Source Between Rooms

Choose a source room, and at least one destination room.  The current selected source in the source room is set as the active source for the destination room(s).  If the “Allow Power Off” option is set to On, if the source room is off the destination rooms will also be turned off, otherwise they will be left alone. If the current source in the room is a Control4 Digital Media native music driver, the destination rooms will be added to that session rather than reselecting the music driver in the rooms.

### Control Button Link

Emulates a tap (sends DO_PUSH then DO_CLICK), press (sends DO_PUSH) or release (sends DO_RELEASE) of a specified button (select from the …) list.

## Troubleshooting

Most issues will be around incorrect BUTTON_LINK connections or attempting to program on events and variables or use BUTTON_LINK connections that aren’t available when multiple rooms are selected.  It is possible to have multiple copies of this driver each controlling overlapping groups of rooms – e.g Entire House, Ground Floor, Living Space, Living Room – and this capability should be used to target exact functions from BUTTON_LINK and programming as desired.

If a BUTTON_LINK doesn’t have any effect on a specific device, it is likely that the device does not implement the named command.  This can be easily fixed for an IR c4i driver (edit the driver in Driver Wizard and add the PLAYPAUSE command) but is less easily fixed for IP or c4z drivers.
The most common “missing” command is likely to be PLAYPAUSE; this is implemented on most media player devices (and on any native Digital Audio source), but many cable/satellite boxes are likely to be missing it.

If you encounter odd behavior with room volume, make sure that there is only a default volume set either in the Room Control driver (and only one!) **OR** the Room driver itself.  Setting a default volume in more than one place for any one room will cause unexpected and unpredictable behavior.

## Change Log

- 139

DRIV-9721 - Correctly setup Listen Default Volume variable

- 138

DRIV-5339, DRIV-9650 - Direct Preset Selection button RFPs now check for DO_PUSH

- 137

DRIV-5173 - Catch infinite loop on reverse cyclers when no presets
DRIV-9579 - Set lights variables on startup

- 136

DRIV-8246 - Add Variables for Any Lights On and All Lights On

- 135

DRIV-9113, DRIV-7957, DRIV-5258 - Allow for reselect of a preset that was the last one to be selected, as long as the primary room does not currently have that preset source selected.

- 134

DRIV-7587 - watch media DB for items being added/removed, and update media property lists accordingly

- 133

DRIV-2985 - Add writable variables for Default Watch and Default Listen Volumes
DRIV-6596 - Add semaphore to guard against firing events during OnPIP

- 132

Add variable for room ID(s) controlled by the driver

- 130

DRIV-4567- When first adding to a project, auto rename based on the room the driver is added to.

- 129

DRIV-4468- Fix Refresh Navigators re-triggering Room On/Off events

- 128

DRIV-4329 - Fix PLAYPAUSE binding for Spotify Connect

- 127

Resolve issue where if multiple rooms were selected, if more than one room had the same device selected then commands would be sent multiple times to that device

- 126

AV-4861 – Selecting a Digital Media source as a preset with no media DB item selected causes session not to be created

- 125

38394 – Left Watch Mode not firing

38395 – Sync rooms when OSD “active” device doesn’t sync audio if OSD source not shared

Added programming commands to trigger all keypad bindings (dynamic list)

- 124

Reset Cycler to 0 on Room Off always

- 123

Fix preset event fire

Fix preset media persistence

Don't send deselect on media selection

- 122

Correctly reset cycler position if broadcast audio selected elsewhere (also fires preset events)

Add events for preset 1-5 firing to allow for LED color selection

- 121

Add option for media DB items on sources that have broadcast audio entries in media DB

- 119

Add option for ignore OSD

Add connection for cycler (reverse)

- 118

Update session joining logic

- 117

Failover watch source to listen if not available

- 116

Add source sync command

Make PLAYPAUSE multi-room compatible

- 115

Room Turned On event now fires when the room turns on

- 114

Resolve issue where if room was a slave zone on Digital Media, the current device wouldn’t be correct

- 113

Add events for Room Turned On, Left Watch Mode, Left Listen Mode

- 112

Correctly set variables on startup

Change Preset Cycler Reset options

- 111

Add programming commands for non-room commands

Catch Digital Media sources better for events/variables

Fix PLAYPAUSE on DM source not working

- 110

Add PlayPause command

Add Preset Source selection and cycler

Add AV and Lights Off BUTTON_LINK

Correct behavior of variables

Remove auto-rename driver

Add Watch and Listen default volume

Add multiple rooms support

- 108

Fix 33185 – Toggle Lights documentation corrected

- 105

Detect Digital Media devices to trigger events/variables on service change

- 103

Better handling of room variable change

- 102

Remove EV Auto Update code

- 100

Initial release
